# Copyright 2015 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Elasticsearch datastore."""
from __future__ import unicode_literals

from collections import Counter
import codecs
import json
import logging
import socket

from uuid import uuid4

import six

from dateutil import parser, relativedelta
from elasticsearch import Elasticsearch
from elasticsearch.exceptions import ConnectionTimeout
from elasticsearch.exceptions import NotFoundError
from elasticsearch.exceptions import RequestError
# pylint: disable=redefined-builtin
from elasticsearch.exceptions import ConnectionError
from flask import abort

from timesketch.lib.definitions import HTTP_STATUS_CODE_NOT_FOUND

# Setup logging
es_logger = logging.getLogger('timesketch.elasticsearch')
es_logger.setLevel(logging.WARNING)

UPDATE_LABEL_SCRIPT = """
if (ctx._source.timesketch_label == null) {
    ctx._source.timesketch_label = new ArrayList()
}
if (params.remove == true) {
    ctx._source.timesketch_label.removeIf(label -> label.name == params.timesketch_label.name && label.sketch_id == params.timesketch_label.sketch_id);
} else {
    if( ! ctx._source.timesketch_label.contains (params.timesketch_label)) {
        ctx._source.timesketch_label.add(params.timesketch_label)
    }
}
"""

TOGGLE_LABEL_SCRIPT = """
if (ctx._source.timesketch_label == null) {
    ctx._source.timesketch_label = new ArrayList()
}
boolean removedLabel = ctx._source.timesketch_label.removeIf(label -> label.name == params.timesketch_label.name && label.sketch_id == params.timesketch_label.sketch_id);
if (!removedLabel) {
    ctx._source.timesketch_label.add(params.timesketch_label)
}
"""


class ElasticsearchDataStore(object):
    """Implements the datastore."""

    # Number of events to queue up when bulk inserting events.
    DEFAULT_FLUSH_INTERVAL = 1000
    DEFAULT_SIZE = 100
    DEFAULT_LIMIT = DEFAULT_SIZE  # Max events to return
    DEFAULT_FROM = 0
    DEFAULT_STREAM_LIMIT = 5000 # Max events to return when streaming results

    def __init__(self, host='127.0.0.1', port=9200):
        """Create a Elasticsearch client."""
        super(ElasticsearchDataStore, self).__init__()
        self._error_container = {}
        self.client = Elasticsearch([{'host': host, 'port': port}])
        self.import_counter = Counter()
        self.import_events = []

    @staticmethod
    def _build_labels_query(sketch_id, labels):
        """Build Elasticsearch query for Timesketch labels.

        Args:
            sketch_id: Integer of sketch primary key.
            labels: List of label names.

        Returns:
            Elasticsearch query as a dictionary.
        """
        label_query = {
            'bool': {
                'must': []
            }
        }

        for label in labels:
            nested_query = {
                'nested': {
                    'query': {
                        'bool': {
                            'must': [{
                                'term': {
                                    'timesketch_label.name.keyword': label
                                }
                            }, {
                                'term': {
                                    'timesketch_label.sketch_id': sketch_id
                                }
                            }]
                        }
                    },
                    'path': 'timesketch_label'
                }
            }
            label_query['bool']['must'].append(nested_query)
        return label_query

    @staticmethod
    def _build_events_query(events):
        """Build Elasticsearch query for one or more document ids.

        Args:
            events: List of Elasticsearch document IDs.

        Returns:
            Elasticsearch query as a dictionary.
        """
        events_list = [event['event_id'] for event in events]
        query_dict = {'query': {'ids': {'values': events_list}}}
        return query_dict

    @staticmethod
    def _convert_to_time_range(interval):
        """Convert an interval timestamp into start and end dates.

        Args:
            interval: Time frame representation

        Returns:
            Start timestamp in string format.
            End timestamp in string format.
        """
        # return ('2018-12-05T00:00:00', '2018-12-05T23:59:59')
        TS_FORMAT = '%Y-%m-%dT%H:%M:%S'
        get_digits = lambda s: int(''.join(filter(str.isdigit, s)))
        get_alpha = lambda s: ''.join(filter(str.isalpha, s))

        ts_parts = interval.split(' ')
        # The start date could be 1 or 2 first items
        start = ' '.join(ts_parts[0:len(ts_parts)-2])
        minus = get_digits(ts_parts[-2])
        plus = get_digits(ts_parts[-1])
        interval = get_alpha(ts_parts[-1])

        start_ts = parser.parse(start)

        rd = relativedelta.relativedelta
        if interval == 's':
            start_range = start_ts - rd(seconds=minus)
            end_range = start_ts + rd(seconds=plus)
        elif interval == 'm':
            start_range = start_ts - rd(minutes=minus)
            end_range = start_ts + rd(minutes=plus)
        elif interval == 'h':
            start_range = start_ts - rd(hours=minus)
            end_range = start_ts + rd(hours=plus)
        elif interval == 'd':
            start_range = start_ts - rd(days=minus)
            end_range = start_ts + rd(days=plus)
        else:
            raise RuntimeError('Unable to parse the timestamp: '
                               + str(interval))

        return start_range.strftime(TS_FORMAT), end_range.strftime(TS_FORMAT)

    def build_query(self, sketch_id, query_string, query_filter, query_dsl=None,
                    aggregations=None):
        """Build Elasticsearch DSL query.

        Args:
            sketch_id: Integer of sketch primary key
            query_string: Query string
            query_filter: Dictionary containing filters to apply
            query_dsl: Dictionary containing Elasticsearch DSL query
            aggregations: Dict of Elasticsearch aggregations

        Returns:
            Elasticsearch DSL query as a dictionary
        """

        if query_dsl:
            if not isinstance(query_dsl, dict):
                query_dsl = json.loads(query_dsl)
            # Remove any aggregation coming from user supplied Query DSL.
            # We have no way to display this data in a good way today.
            if query_dsl.get('aggregations', None):
                del query_dsl['aggregations']
            return query_dsl

        if query_filter.get('events', None):
            events = query_filter['events']
            return self._build_events_query(events)

        query_dsl = {
            'query': {
                'bool': {
                    'must': [],
                    'must_not': [],
                    'filter': []
                }
            }
        }

        # TODO: Remove when old UI has been deprecated.
        if query_filter.get('star', None):
            label_query = self._build_labels_query(sketch_id, ['__ts_star'])
            query_string = '*'
            query_dsl['query']['bool']['must'].append(label_query)

        # TODO: Remove when old UI has been deprecated.
        if query_filter.get('time_start', None):
            query_dsl['query']['bool']['filter'] = [{
                'bool': {
                    'should': [{
                        'range': {
                            'datetime': {
                                'gte': query_filter['time_start'],
                                'lte': query_filter['time_end']
                            }
                        }
                    }]
                }
            }]

        if query_string:
            query_dsl['query']['bool']['must'].append(
                {'query_string': {'query': query_string}})

        # New UI filters
        if query_filter.get('chips', None):
            labels = []
            must_filters = query_dsl['query']['bool']['must']
            must_not_filters = query_dsl['query']['bool']['must_not']
            datetime_ranges = {
                'bool': {
                    'should': [],
                    'minimum_should_match': 1
                }
            }

            for chip in query_filter['chips']:
                # Exclude chips that the user disabled
                if not chip.get('active', True):
                    continue

                if chip['type'] == 'label':
                    labels.append(chip['value'])

                elif chip['type'] == 'term':
                    term_filter = {
                        'match_phrase': {
                            '{}'.format(chip['field']): {
                                'query': "{}".format(chip['value'])
                            }
                        }
                    }

                    if chip['operator'] == 'must':
                        must_filters.append(term_filter)

                    elif chip['operator'] == 'must_not':
                        must_not_filters.append(term_filter)

                elif chip['type'].startswith('datetime'):
                    range_filter = lambda start, end: {
                        'range': {
                            'datetime': {
                                'gte': start,
                                'lte': end
                            }
                        }
                    }
                    if chip['type'] == 'datetime_range':
                        start, end = chip['value'].split(',')
                    elif chip['type'] == 'datetime_interval':
                        start, end = self._convert_to_time_range(chip['value'])
                    else:
                        continue
                    datetime_ranges['bool']['should'].append(
                        range_filter(start, end))

            label_filter = self._build_labels_query(sketch_id, labels)
            must_filters.append(label_filter)
            must_filters.append(datetime_ranges)

        # Pagination
        if query_filter.get('from', None):
            query_dsl['from'] = query_filter['from']

        # Number of events to return
        if query_filter.get('size', None):
            query_dsl['size'] = query_filter['size']

        # Make sure we are sorting.
        if not query_dsl.get('sort', None):
            query_dsl['sort'] = {
                'datetime': query_filter.get('order', 'asc')
            }

        # Add any pre defined aggregations
        if aggregations:
            # post_filter happens after aggregation so we need to move the
            # filter to the query instead.
            if query_dsl.get('post_filter', None):
                query_dsl['query']['bool']['filter'] = query_dsl[
                    'post_filter']
                query_dsl.pop('post_filter', None)
            query_dsl['aggregations'] = aggregations

        return query_dsl

    def search(self, sketch_id, query_string, query_filter, query_dsl, indices,
               count=False, aggregations=None, return_fields=None,
               enable_scroll=False):
        """Search ElasticSearch. This will take a query string from the UI
        together with a filter definition. Based on this it will execute the
        search request on ElasticSearch and get result back.

        Args:
            sketch_id: Integer of sketch primary key
            query_string: Query string
            query_filter: Dictionary containing filters to apply
            query_dsl: Dictionary containing Elasticsearch DSL query
            indices: List of indices to query
            count: Boolean indicating if we should only return result count
            aggregations: Dict of Elasticsearch aggregations
            return_fields: List of fields to return
            enable_scroll: If Elasticsearch scroll API should be used

        Returns:
            Set of event documents in JSON format
        """

        scroll_timeout = None
        if enable_scroll:
            scroll_timeout = '1m'  # Default to 1 minute scroll timeout

        # Exit early if we have no indices to query
        if not indices:
            return {'hits': {'hits': [], 'total': 0}, 'took': 0}

        # Check if we have specific events to fetch and get indices.
        if query_filter.get('events', None):
            indices = {
                event['index']
                for event in query_filter['events']
                if event['index'] in indices
            }

        query_dsl = self.build_query(sketch_id, query_string, query_filter,
                                     query_dsl, aggregations)

        # Default search type for elasticsearch is query_then_fetch.
        search_type = 'query_then_fetch'

        # Only return how many documents matches the query.
        if count:
            del query_dsl['sort']
            count_result = self.client.count(
                body=query_dsl, index=list(indices))
            return count_result.get('count', 0)

        if not return_fields:
            # Suppress the lint error because elasticsearch-py adds parameters
            # to the function with a decorator and this makes pylint sad.
            # pylint: disable=unexpected-keyword-arg
            return self.client.search(
                body=query_dsl,
                index=list(indices),
                search_type=search_type,
                scroll=scroll_timeout)

        # The argument " _source_include" changed to "_source_includes" in
        # ES version 7. This check add support for both version 6 and 7 clients.
        # pylint: disable=unexpected-keyword-arg
        try:
            if self.version.startswith('6'):
                _search_result = self.client.search(
                    body=query_dsl,
                    index=list(indices),
                    search_type=search_type,
                    _source_include=return_fields,
                    scroll=scroll_timeout)
            else:
                _search_result = self.client.search(
                    body=query_dsl,
                    index=list(indices),
                    search_type=search_type,
                    _source_includes=return_fields,
                    scroll=scroll_timeout)
        except RequestError as e:
            root_cause = e.info.get('error', {}).get('root_cause')
            if root_cause:
                error_items = []
                for cause in root_cause:
                    error_items.append(
                        '[{0:s}] {1:s}'.format(
                            cause.get('type', ''), cause.get('reason', '')))
                cause = ', '.join(error_items)
            else:
                cause = str(e)

            es_logger.error(
                'Unable to run search query: {0:s}'.format(cause),
                exc_info=True)
            raise ValueError(cause)

        return _search_result

    def search_stream(self, sketch_id=None, query_string=None,
                      query_filter=None, query_dsl=None, indices=None,
                      return_fields=None, enable_scroll=True):
        """Search ElasticSearch. This will take a query string from the UI
        together with a filter definition. Based on this it will execute the
        search request on ElasticSearch and get result back.

        Args :
            sketch_id: Integer of sketch primary key
            query_string: Query string
            query_filter: Dictionary containing filters to apply
            query_dsl: Dictionary containing Elasticsearch DSL query
            indices: List of indices to query
            return_fields: List of fields to return
            enable_scroll: Boolean determing whether scrolling is enabled.

        Returns:
            Generator of event documents in JSON format
        """

        if not query_filter.get('size'):
            query_filter['size'] = self.DEFAULT_STREAM_LIMIT

        if not query_filter.get('terminate_after'):
            query_filter['terminate_after'] = self.DEFAULT_STREAM_LIMIT

        result = self.search(
            sketch_id=sketch_id,
            query_string=query_string,
            query_dsl=query_dsl,
            query_filter=query_filter,
            indices=indices,
            return_fields=return_fields,
            enable_scroll=enable_scroll)

        if enable_scroll:
            scroll_id = result['_scroll_id']
            scroll_size = result['hits']['total']
        else:
            scroll_id = None
            scroll_size = 0

        # Elasticsearch version 7.x returns total hits as a dictionary.
        # TODO: Refactor when version 6.x has been deprecated.
        if isinstance(scroll_size, dict):
            scroll_size = scroll_size.get('value', 0)

        for event in result['hits']['hits']:
            yield event

        while scroll_size > 0:
            # pylint: disable=unexpected-keyword-arg
            result = self.client.scroll(scroll_id=scroll_id, scroll='5m')
            scroll_id = result['_scroll_id']
            scroll_size = len(result['hits']['hits'])
            for event in result['hits']['hits']:
                yield event

    def get_filter_labels(self, sketch_id, indices):
        """Aggregate labels for a sketch.

        Args:
            sketch_id: The Sketch ID
            indices: List of indices to aggregate on

        Returns:
            List with label names.
        """
        # This is a workaround to return all labels by setting the max buckets
        # to something big. If a sketch has more than this amount of labels
        # the list will be incomplete but it should be uncommon to have >10k
        # labels in a sketch.
        max_labels = 10000

        # pylint: disable=line-too-long
        aggregation = {
            'aggs': {
                'nested': {
                    'nested': {
                        'path': 'timesketch_label'
                    },
                    'aggs': {
                        'inner': {
                            'filter': {
                                'bool': {
                                    'must': [{
                                        'term': {
                                            'timesketch_label.sketch_id': sketch_id
                                        }
                                    }]
                                }
                            },
                            'aggs': {
                                'labels': {
                                    'terms': {
                                        'size': max_labels,
                                        'field': 'timesketch_label.name.keyword'
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }

        labels = []
        # pylint: disable=unexpected-keyword-arg
        result = self.client.search(index=indices, body=aggregation, size=0)
        buckets = result.get(
            'aggregations', {}).get('nested', {}).get('inner', {}).get(
                'labels', {}).get('buckets', [])
        for bucket in buckets:
            # Filter out special labels like __ts_star etc.
            if bucket['key'].startswith('__'):
                continue
            labels.append(bucket['key'])
        return labels

    def get_event(self, searchindex_id, event_id):
        """Get one event from the datastore.

        Args:
            searchindex_id: String of ElasticSearch index id
            event_id: String of ElasticSearch event id

        Returns:
            Event document in JSON format
        """
        try:
            # Suppress the lint error because elasticsearch-py adds parameters
            # to the function with a decorator and this makes pylint sad.
            # pylint: disable=unexpected-keyword-arg
            if self.version.startswith('6'):
                event = self.client.get(
                    index=searchindex_id,
                    id=event_id,
                    doc_type='_all',
                    _source_exclude=['timesketch_label'])
            else:
                event = self.client.get(
                    index=searchindex_id,
                    id=event_id,
                    doc_type='_all',
                    _source_excludes=['timesketch_label'])

            return event

        except NotFoundError:
            abort(HTTP_STATUS_CODE_NOT_FOUND)

    def count(self, indices):
        """Count number of documents.

        Args:
            indices: List of indices.

        Returns:
            Number of documents.
        """
        if not indices:
            return 0
        try:
            result = self.client.count(index=indices)
        except (NotFoundError, RequestError):
            es_logger.error(
                'Unable to count indexes (index not found)',
                exc_info=True)
            return 0
        return result.get('count', 0)

    def set_label(self, searchindex_id, event_id, event_type, sketch_id,
                  user_id, label, toggle=False, remove=False,
                  single_update=True):
        """Set label on event in the datastore.

        Args:
            searchindex_id: String of ElasticSearch index id
            event_id: String of ElasticSearch event id
            event_type: String of ElasticSearch document type
            sketch_id: Integer of sketch primary key
            user_id: Integer of user primary key
            label: String with the name of the label
            remove: Optional boolean value if the label should be removed
            toggle: Optional boolean value if the label should be toggled
            single_update: Boolean if the label should be indexed immediately.

        Returns:
            Dict with updated document body, or None if this is a single update.
        """
        # Elasticsearch painless script.
        update_body = {
            'script': {
                'lang': 'painless',
                'source': UPDATE_LABEL_SCRIPT,
                'params': {
                    'timesketch_label': {
                        'name': str(label),
                        'user_id': user_id,
                        'sketch_id': sketch_id
                    },
                    remove: remove
                }
            }
        }

        if toggle:
            update_body['script']['source'] = TOGGLE_LABEL_SCRIPT

        if not single_update:
            script = update_body['script']
            return dict(
                source=script['source'], lang=script['lang'],
                params=script['params']
            )

        doc = self.client.get(
            index=searchindex_id, id=event_id, doc_type='_all')
        try:
            doc['_source']['timesketch_label']
        except KeyError:
            doc = {'doc': {'timesketch_label': []}}
            self.client.update(
                index=searchindex_id,
                doc_type=event_type,
                id=event_id,
                body=doc)

        self.client.update(
            index=searchindex_id,
            id=event_id,
            doc_type=event_type,
            body=update_body)

        return None

    def create_index(self, index_name=uuid4().hex, doc_type='generic_event'):
        """Create index with Timesketch settings.

        Args:
            index_name: Name of the index. Default is a generated UUID.
            doc_type: Name of the document type. Default id generic_event.

        Returns:
            Index name in string format.
            Document type in string format.
        """
        _document_mapping = {
            'properties': {
                'timesketch_label': {
                    'type': 'nested'
                },
                'datetime': {
                    'type': 'date'
                }
            }
        }

        # TODO: Remove when we deprecate Elasticsearch version 6.x
        if self.version.startswith('6'):
            _document_mapping = {doc_type: _document_mapping}

        if not self.client.indices.exists(index_name):
            try:
                self.client.indices.create(
                    index=index_name, body={'mappings': _document_mapping})
            except ConnectionError:
                raise RuntimeError('Unable to connect to Timesketch backend.')
            except RequestError:
                index_exists = self.client.indices.exists(index_name)
                es_logger.warning(
                    'Attempting to create an index that already exists '
                    '({0:s} - {1:s})'.format(index_name, str(index_exists)))

        # We want to return unicode here to keep SQLalchemy happy.
        if six.PY2:
            if not isinstance(index_name, six.text_type):
                index_name = codecs.decode(index_name, 'utf-8')

            if not isinstance(doc_type, six.text_type):
                doc_type = codecs.decode(doc_type, 'utf-8')

        return index_name, doc_type

    def delete_index(self, index_name):
        """Delete Elasticsearch index.

        Args:
            index_name: Name of the index to delete.
        """
        if self.client.indices.exists(index_name):
            try:
                self.client.indices.delete(index=index_name)
            except ConnectionError as e:
                raise RuntimeError(
                    'Unable to connect to Timesketch backend: {}'.format(e)
                )

    def import_event(self, index_name, event_type, event=None, event_id=None,
                     flush_interval=DEFAULT_FLUSH_INTERVAL):
        """Add event to Elasticsearch.

        Args:
            index_name: Name of the index in Elasticsearch
            event_type: Type of event (e.g. plaso_event)
            event: Event dictionary
            event_id: Event Elasticsearch ID
            flush_interval: Number of events to queue up before indexing
        """
        if event:
            for k, v in event.items():
                if not isinstance(k, six.text_type):
                    k = codecs.decode(k, 'utf8')

                # Make sure we have decoded strings in the event dict.
                if isinstance(v, six.binary_type):
                    v = codecs.decode(v, 'utf8')

                event[k] = v

            # Header needed by Elasticsearch when bulk inserting.
            header = {
                'index': {
                    '_index': index_name,
                }
            }
            update_header = {
                'update': {
                    '_index': index_name,
                    '_id': event_id
                }
            }

            # TODO: Remove when we deprecate Elasticsearch version 6.x
            if self.version.startswith('6'):
                header['index']['_type'] = event_type
                update_header['update']['_type'] = event_type

            if event_id:
                # Event has "lang" defined if there is a script used for import.
                if event.get('lang'):
                    event = {'script': event}
                else:
                    event = {'doc': event}
                header = update_header

            self.import_events.append(header)
            self.import_events.append(event)
            self.import_counter['events'] += 1

            if self.import_counter['events'] % int(flush_interval) == 0:
                _ = self.flush_queued_events()
                self.import_events = []
        else:
            # Import the remaining events in the queue.
            if self.import_events:
                _ = self.flush_queued_events()

        return self.import_counter['events']

    def flush_queued_events(self):
        """Flush all queued events.

        Returns:
            dict: A dict object that contains the number of events
                that were sent to Elastic as well as information
                on whether there were any errors, and what the
                details of these errors if any.
        """
        if not self.import_events:
            return {}

        return_dict = {
            'number_of_events': len(self.import_events) / 2,
            'total_events': self.import_counter['events'],
        }

        try:
            results = self.client.bulk(body=self.import_events)
        except (ConnectionTimeout, socket.timeout):
            # TODO: Add a retry here.
            es_logger.error('Unable to add events', exc_info=True)

        errors_in_upload = results.get('errors', False)
        return_dict['errors_in_upload'] = errors_in_upload

        if errors_in_upload:
            items = results.get('items', [])
            return_dict['errors'] = []

            es_logger.error('Errors while attempting to upload events.')
            for item in items:
                index = item.get('index', {})
                index_name = index.get('_index', 'N/A')

                _ = self._error_container.setdefault(
                    index_name, {
                        'errors': [],
                        'types': Counter(),
                        'details': Counter()
                    }
                )

                error_counter = self._error_container[index_name]['types']
                error_detail_counter = self._error_container[index_name][
                    'details']
                error_list = self._error_container[index_name]['errors']

                error = index.get('error', {})
                status_code = index.get('status', 0)
                doc_id = index.get('_id', '')
                caused_by = error.get('caused_by', {})

                caused_reason = caused_by.get(
                    'reason', 'Unkown Detailed Reason')

                error_counter[error.get('type')] += 1
                detail_msg = '{0:s}/{1:s}'.format(
                    caused_by.get('type', 'Unknown Detailed Type'),
                    ' '.join(caused_reason.split()[:5])
                )
                error_detail_counter[detail_msg] += 1

                error_msg = '<{0:s}> {1:s} [{2:s}/{3:s}]'.format(
                    error.get('type', 'Unknown Type'),
                    error.get('reason', 'No reason given'),
                    caused_by.get('type', 'Unknown Type'),
                    caused_reason,
                )
                error_list.append(error_msg)
                es_logger.error(
                    'Unable to upload document: {0:s} to index {1:s} - '
                    '[{2:d}] {3:s}'.format(
                        doc_id, index_name, status_code, error_msg))

        return_dict['error_container'] = self._error_container

        self.import_events = []
        return return_dict

    @property
    def version(self):
        """Get Elasticsearch version.

        Returns:
          Version number as a string.
        """
        version_info = self.client.info().get('version')
        return version_info.get('number')



"""Tests for TaggerSketchPlugin."""
import yaml

from timesketch.lib import emojis
from timesketch.lib.testlib import BaseTest


class TestTaggerPlugin(BaseTest):
    """Tests the functionality of the analyzer."""

    def _config_validation(self, config):
        """Validate that all items of a config are valid."""
        query = config.get('query_string', config.get('query_dsl'))
        self.assertIsNotNone(query)
        self.assertIsInstance(query, str)

        emojis_to_add = config.get('emojis')
        if emojis_to_add:
            self.assertIsInstance(emojis_to_add, (list, tuple))
            for emoji_name in emojis_to_add:
                emoji_code = emojis.get_emoji(emoji_name)
                self.assertNotEqual(emoji_code, '')

        tags = config.get('tags')
        if tags:
            self.assertIsInstance(tags, (list, tuple))

        create_view = config.get('create_view')
        if create_view:
            self.assertIsInstance(create_view, bool)

    # TODO: Add tests for the tagger.
    def test_config(self):
        """Tests that the config is valid."""
        test_config = yaml.safe_load("""
        place_holder:
          query_string: '*'
          tags: ['place-holder']
          emojis: ['ID_BUTTON']
          """)

        self.assertIsInstance(test_config, dict)

        for key, value in iter(test_config.items()):
            self.assertIsInstance(key, str)
            self.assertIsInstance(value, dict)
            self._config_validation(value)



# Copyright 2015 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Elasticsearch aggregations."""

from __future__ import unicode_literals

# TODO: Port to new aggregation framework.

# Max number of search hits to chart
# TODO: Move to config
MAX_RESULT_LIMIT = 300000


def heatmap(es_client, sketch_id, query_string, query_filter, query_dsl,
            indices):
    """Aggregate query results into number of events per hour/day.

    Args:
        es_client: Elasticsearch client (instance of ElasticSearchDatastore)
        sketch_id: Integer of sketch primary key
        query_string: Query string
        query_filter: Dictionary containing filters to apply
        query_dsl: Dictionary containing Elasticsearch DSL to apply
        indices: List of indices to query

    returns:
        List of events per hour/day
    """

    # scripts don't like our pagination method, so remove them
    query_filter.pop("size", None)
    query_filter.pop("from", None)

    result_count = es_client.search(
        sketch_id, query_string, query_filter, query_dsl, indices, count=True)

    # Exit early if we have too many search hits.
    if result_count > MAX_RESULT_LIMIT or result_count == 0:
        return []

    days_map = {
        'Mon': 1,
        'Tue': 2,
        'Wed': 3,
        'Thu': 4,
        'Fri': 5,
        'Sat': 6,
        'Sun': 7,
    }

    # Elasticsearch 5.x doesn't support toString for dates. Use SimpleDateFormat
    # instead.
    # pylint: disable=line-too-long

    if es_client.version.startswith('5.'):
        source_script = 'new SimpleDateFormat(params.format).format(new Date(doc["datetime"].value))'
    else:
        source_script = 'doc["datetime"].value.toString(params.format);'

    aggregation = {
        'byDay': {
            'terms': {
                'script': {
                    'source': source_script,
                    'lang': 'painless',
                    'params': {
                        'format': 'E'
                    }
                }
            },
            'aggs': {
                'byHour': {
                    'terms': {
                        'order': {
                            '_term': 'asc'
                        },
                        'script': {
                            'source': source_script,
                            'lang': 'painless',
                            'params': {
                                'format': 'H'
                            }
                        },
                        'size': 24
                    }
                }
            }
        }
    }

    search_result = es_client.search(
        sketch_id,
        query_string,
        query_filter,
        query_dsl,
        indices,
        aggregations=aggregation,
        return_fields=None,
        enable_scroll=False)

    try:
        aggregation_result = search_result['aggregations']
        day_buckets = aggregation_result['byDay']['buckets']
    except KeyError:
        day_buckets = []

    per_hour = {}
    for day in range(1, 8):
        for hour in range(0, 24):
            per_hour[(day, hour)] = 0

    for day_bucket in day_buckets:
        day = days_map[day_bucket.get('key')]
        day_hours = day_bucket['byHour']['buckets']

        for day_hour in day_hours:
            hour = int(day_hour['key'])
            count = day_hour['doc_count']
            per_hour[(day, int(hour))] = count

    return [dict(day=k[0], hour=k[1], count=v) for k, v in per_hour.items()]


def histogram(es_client, sketch_id, query_string, query_filter, query_dsl,
              indices):
    """Aggregate query results into number of events per time interval.

    Args:
        es_client: Elasticsearch client (instance of ElasticSearchDatastore)
        sketch_id: Integer of sketch primary key
        query_string: Query string
        query_filter: Dictionary containing filters to apply
        query_dsl: Dictionary containing Elasticsearch DSL to apply
        indices: List of indices to query

    returns:
        List of events per hour/day
    """
    # scripts don't like our pagination method, so remove them
    query_filter.pop("size", None)
    query_filter.pop("from", None)

    result_count = es_client.search(
        sketch_id, query_string, query_filter, query_dsl, indices, count=True)

    # Exit early if we have too many search hits.
    if result_count > MAX_RESULT_LIMIT or result_count == 0:
        return []

    # Default aggregation config
    interval = 'day'
    min_doc_count = 10

    if result_count < 10:
        min_doc_count = 1

    aggregation = {
        'histogram': {
            'date_histogram': {
                'min_doc_count': min_doc_count,
                'field': 'datetime',
                'interval': interval,
                'format': 'yyyy-MM-dd'
            }
        }
    }

    if result_count < MAX_RESULT_LIMIT:
        search_result = es_client.search(
            sketch_id,
            query_string,
            query_filter,
            query_dsl,
            indices,
            aggregations=aggregation,
            return_fields=None,
            enable_scroll=False)
    else:
        search_result = {}

    try:
        aggregation_result = search_result['aggregations']
        if aggregation_result.get('exclude', None):
            buckets = aggregation_result['exclude']['histogram']['buckets']
        else:
            buckets = aggregation_result['histogram']['buckets']
    except KeyError:
        buckets = []

    return buckets



"""Sessionizing sketch analyzer plugin."""

from __future__ import unicode_literals

from timesketch.lib.analyzers import interface
from timesketch.lib.analyzers import manager


class SessionizerSketchPlugin(interface.BaseSketchAnalyzer):
    """Sessionizing sketch analyzer. All events in sketch with id sketch_id
    are grouped in sessions based on the time difference between them. Two
    consecutive events are in the same session if the time difference between
    them is less or equal then max_time_diff_micros.

    Attributes:
        NAME (str): The name of the sessionizer.
        max_time_diff_micros (int): The maximum time difference between two
            events in the same session, in microseconds.
        query (str): The Elasticsearch query string query identifying the
            events to be sessionized.
        session_type (str): Used to label the events that are sessionized.
    """

    NAME = 'sessionizer'
    # TODO max_time_diff_micros should be configurable
    max_time_diff_micros = 300000000
    query = '*'
    session_type = 'all_events'

    def run(self):
        """Entry point for the analyzer. Allocates each event a session_id
        attribute.
        Returns:
            String containing the number of sessions created.
        """
        return_fields = ['timestamp']

        # event_stream returns an ordered generator of events (by time)
        # therefore no further sorting is needed.
        events = self.event_stream(query_string=self.query,
                                   return_fields=return_fields)
        session_num = 0

        try:
            first_event = next(events)
            last_timestamp = first_event.source.get('timestamp')
            session_num = 1
            self.annotateEvent(first_event, session_num)

            for event in events:
                curr_timestamp = event.source.get('timestamp')
                if curr_timestamp - last_timestamp > self.max_time_diff_micros:
                    session_num += 1
                self.annotateEvent(event, session_num)
                last_timestamp = curr_timestamp

        except StopIteration:
            pass

        return ('Sessionizing completed, number of session created:'
                ' {0:d}'.format(session_num))

    def annotateEvent(self, event, session_num):
        """Annotate an event with a session ID. Store IDs as dictionary entries
        corresponding to the type of session.
        Args:
            event: The event to annotate.
            session_num: The session ID.
        """
        current_attr = event.source.get('session_id')
        if current_attr is None:
            event.add_attributes({'session_id': {self.session_type:
                                                 session_num}})
        else:
            current_attr[self.session_type] = session_num
            event.add_attributes({'session_id': current_attr})

        event.commit()

manager.AnalysisManager.register_analyzer(SessionizerSketchPlugin)



"""Sequenced activity sessionizing sketch analyzer plugin."""

from __future__ import unicode_literals

from timesketch.lib.analyzers import sessionizer


class SequenceSessionizerSketchPlugin(sessionizer.SessionizerSketchPlugin):
    """Base class for specialized sequenced activity sessionizings
    sketch plugins.

    Attributes:
        event_seq: List of dictionaries of attributes describing the events.
        event_storage: List of events.
        num_event_to_find: Number that shows which event from the sequence
            should be matched.
        recording: Shows if currently are finded and stored events that match
            the specified sequence of events.
        session_num: Counter for the number of sessions.
    """
    event_seq = []
    event_storage = []
    num_event_to_find = 0
    recording = False
    return_fields = ['timestamp']
    session_num = 0
    session_type = None

    def run(self):
        """Entry point for the analyzer.

        Allocates each event between the first event of the event_seq and the
        last event of the event_seq an session_type attribute with a value
        session_num.

        Returns:
            String containing the name of the event sequence and the
            number of sessions created.
        """
        if self.session_type is None or self.session_type == '':
            raise ValueError('No session_type provided.')
        if self.event_seq is None or self.event_seq == []:
            raise ValueError('No event_seq provided.')
        # If return_fields in None, then all attributes are provided.
        if self.return_fields is not None:
            self.build_return_fields()

        # event_stream returns an ordered generator of events (by time)
        # therefore no further sorting is needed.
        events = self.event_stream(query_string=self.query,
                                   return_fields=self.return_fields)

        last_timestamp = None
        for event in events:
            curr_timestamp = event.source.get('timestamp')
            if last_timestamp and \
                (curr_timestamp - last_timestamp > self.max_time_diff_micros):
                self.flush_events(drop=True)
            self.process_event(event)
            last_timestamp = curr_timestamp

        self.sketch.add_view('Session view',
                             self.NAME,
                             query_string='session_id.{0:s}:*'.format(
                                 self.session_type))

        return ('Sessionizing completed, number of {0:s} sessions created:'
                ' {1:d}'.format(self.session_type, self.session_num))

    def process_event(self, event):
        """Process event depending on if the event is significant for the
        searched event sequence.

        Event is significant if it matches the current event in the event_seq
        or if in the timeline it is between two consistent events from the
        event_seq.

        Args:
            event: Event to process.
        """
        if self.match_event(event):
            self.recording = True
            self.event_storage.append(event)
            self.num_event_to_find += 1

            if self.num_event_to_find == len(self.event_seq):
                self.flush_events()
            return

        if self.recording:
            self.event_storage.append(event)

    def flush_events(self, drop=False):
        """Annotates or clears events in event_storage according to the flag
        drop and resets session recording state.

        Args:
            drop: If True, the stored events will not be committed to the
                database.
        """
        if not drop:
            self.session_num += 1
            for event in self.event_storage:
                self.annotateEvent(event, self.session_num)

        self.recording = False
        self.num_event_to_find = 0
        self.event_storage = []

    def match_event(self, event):
        """Compare event with the event to search for in the event sequence.

        Args:
            event: Event to compare with the one in the event sequence.

        Returns:
            If event matches the event to search for in the event sequence.
        """
        event_to_match = self.event_seq[self.num_event_to_find]

        for key, value in event_to_match.items():
            if key not in event.source.keys():
                return False
            if value not in event.source.get(key):
                return False
        return True

    def build_return_fields(self):
        """Add missing fields to return_fields. Additional fields are not
        removed."""
        if 'timestamp' not in self.return_fields:
            self.return_fields.append('timestamp')
        for event in self.event_seq:
            self.return_fields.extend(event)
        self.return_fields = list(set(self.return_fields))



# Copyright 2019 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for analysis utils."""

from __future__ import unicode_literals

import six

import pandas as pd

from timesketch.lib.testlib import BaseTest
from timesketch.lib.analyzers import utils


class TestAnalyzerUtils(BaseTest):
    """Tests the functionality of the utilities."""

    def test_get_domain_from_url(self):
        """Test get_domain_from_url function."""
        url = 'http://www.example.com/?foo=bar'
        domain = utils.get_domain_from_url(url)
        self.assertEqual(domain, 'www.example.com')

    def test_get_tld_from_domain(self):
        """Test get_tld_from_domain function."""
        domain = 'this.is.a.subdomain.example.com'
        tld = utils.get_tld_from_domain(domain)
        self.assertEqual(tld, 'example.com')

        domain = 'a'
        tld = utils.get_tld_from_domain(domain)
        self.assertEqual(tld, 'a')

        domain = 'example.com'
        tld = utils.get_tld_from_domain(domain)
        self.assertEqual(tld, 'example.com')

    def test_strip_www_from_domain(self):
        """Test strip_www_from_domain function."""
        domain = 'www.mbl.is'
        stripped = utils.strip_www_from_domain(domain)
        self.assertEqual(stripped, 'mbl.is')

        domain = 'mbl.is'
        stripped = utils.strip_www_from_domain(domain)
        self.assertEqual(stripped, domain)

    def test_get_cdn_provider(self):
        """Test get_cdn_provider function."""
        domain = 'foobar.gstatic.com'
        provider = utils.get_cdn_provider(domain)
        self.assertIsInstance(provider, six.text_type)
        self.assertEqual(provider, 'Google')

        domain = 'www.mbl.is'
        provider = utils.get_cdn_provider(domain)
        self.assertIsInstance(provider, six.text_type)
        self.assertEqual(provider, '')

    def test_get_events_from_data_frame(self):
        """Test getting all events from data frame."""
        lines = [
            {'_id': '123', '_type': 'manual', '_index': 'asdfasdf',
             'tool': 'isskeid'},
            {'_id': '124', '_type': 'manual', '_index': 'asdfasdf',
             'tool': 'tong'},
            {'_id': '125', '_type': 'manual', '_index': 'asdfasdf',
             'tool': 'klemma'},
        ]
        frame = pd.DataFrame(lines)

        events = list(utils.get_events_from_data_frame(frame, None))
        self.assertEqual(len(events), 3)
        ids = [x.event_id for x in events]
        self.assertEqual(set(ids), set(['123', '124', '125']))



# Copyright 2020 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module holds methods and classes to export events."""

from __future__ import unicode_literals

import io
import json
import logging

import pandas as pd

from timesketch.api.v1 import utils
from timesketch.lib.stories import api_fetcher as story_api_fetcher


logger = logging.getLogger('timesketch.api_exporter')


def export_aggregation(aggregation, sketch, zip_file):
    """Export an aggregation from a sketch and write it to a ZIP file.

    Args:
        aggregation (timesketch.models.sketch.Aggregation): an aggregation
            object.
        sketch (timesketch.models.sketch.Sketch): a sketch object.
        zip_file (ZipFile): a zip file handle that can be used to write
            content to.
    """
    name = '{0:04d}_{1:s}'.format(aggregation.id, aggregation.name)
    parameters = json.loads(aggregation.parameters)
    result_obj, meta = utils.run_aggregator(
        sketch.id, aggregator_name=aggregation.agg_type,
        aggregator_parameters=parameters)

    zip_file.writestr(
        'aggregations/{0:s}.meta'.format(name), data=json.dumps(meta))

    html = result_obj.to_chart(
        chart_name=meta.get('chart_type'),
        chart_title=aggregation.name,
        color=meta.get('chart_color'),
        interactive=True, as_html=True)
    zip_file.writestr(
        'aggregations/{0:s}.html'.format(name), data=html)

    string_io = io.StringIO()
    data_frame = result_obj.to_pandas()
    data_frame.to_csv(string_io, index=False)
    string_io.seek(0)
    zip_file.writestr(
        'aggregations/{0:s}.csv'.format(name), data=string_io.read())


def export_aggregation_group(group, sketch, zip_file):
    """Export an aggregation group from a sketch and write it to a ZIP file.

    Args:
        group (timesketch.models.sketch.AggregationGroup): an aggregation
            group object.
        sketch (timesketch.models.sketch.Sketch): a sketch object.
        zip_file (ZipFile): a zip file handle that can be used to write
            content to.
    """
    name = '{0:04d}_{1:s}'.format(group.id, group.name)
    chart, _, meta = utils.run_aggregator_group(group, sketch_id=sketch.id)

    zip_file.writestr(
        'aggregation_groups/{0:s}.meta'.format(name), json.dumps(meta))
    zip_file.writestr(
        'aggregation_groups/{0:s}.html'.format(name), chart.to_html())


def export_story(story, sketch, story_exporter, zip_file):
    """Export a story from a sketch into a ZIP file.

    Args:
        story (timesketch.models.sketch.Story): a story object.
        sketch (timesketch.models.sketch.Sketch): a sketch object.
        story_exporter (timesketch.lib.stories.StoryExporter): an instance of
            a story exporter that can be used to export story content.
        zip_file (ZipFile): a zip file handle that can be used to write
            content to.
    """
    with story_exporter() as exporter:
        data_fetcher = story_api_fetcher.ApiDataFetcher()
        data_fetcher.set_sketch_id(sketch.id)

        exporter.set_data_fetcher(data_fetcher)
        exporter.from_string(story.content)
        exporter.set_creation_date(story.created_at.isoformat())
        if story.user:
            author = story.user.username
        else:
            author = 'System'
        exporter.set_author(author)
        exporter.set_title(story.title)

        zip_file.writestr(
            'stories/{0:04d}_{1:s}.html'.format(
                story.id, story.title),
            data=exporter.export_story())


def query_to_filehandle(
        query_string='', query_dsl='', query_filter=None, sketch=None,
        datastore=None, indices=None):
    """Query the datastore and return back a file object with the results.

    This function takes a query string or DSL, queries the datastore
    and fetches all the events and stores them in a file-like object
    which gets returned back.

    Args:
        query_string (str): Elasticsearch query string.
        query_dsl (str): Elasticsearch query DSL as JSON string.
        query_filter (dict): Filter for the query as a dict.
        sketch (timesketch.models.sketch.Sketch): a sketch object.
        datastore (elastic.ElasticsearchDataStore): the datastore object.
        indices (list): List of indices to query

    Returns:
        file-like object in a CSV format with the results.
    """
    # Ignoring the size limits to reduce the amount of queries
    # needed to get all the data.
    query_filter['terminate_after'] = 10000
    query_filter['size'] = 10000

    if 'from' in query_filter:
        del query_filter['from']

    result = datastore.search(
        sketch_id=sketch.id,
        query_string=query_string,
        query_filter=query_filter,
        query_dsl=query_dsl,
        enable_scroll=True,
        indices=indices)

    scroll_id = result.get('_scroll_id', '')
    if not scroll_id:
        return query_results_to_filehandle(result, sketch)

    data_frame = query_results_to_dataframe(result, sketch)

    total_count = result.get(
        'hits', {}).get('total', {}).get('value', 0)

    if isinstance(total_count, str):
        try:
            total_count = int(total_count, 10)
        except ValueError:
            total_count = 0

    event_count = len(result['hits']['hits'])

    while event_count < total_count:
        # pylint: disable=unexpected-keyword-arg
        result = datastore.client.scroll(
            scroll_id=scroll_id, scroll='1m')
        event_count += len(result['hits']['hits'])
        add_frame = query_results_to_dataframe(result, sketch)
        if add_frame.shape[0]:
            data_frame = pd.concat([data_frame, add_frame], sort=False)
        else:
            logger.warning(
                'Data Frame returned from a search operation was '
                'empty, count {0:d} out of {1:d} total. Query is: '
                '"{2:s}"'.format(
                    event_count, total_count,
                    query_string or query_dsl))

    fh = io.StringIO()
    data_frame.to_csv(fh, index=False)
    fh.seek(0)
    return fh


def query_results_to_dataframe(result, sketch):
    """Returns a data frame from a Elastic query result dict.

    Args:
        result (dict): a dict that contains the response from a
            Elastic datastore search.
        sketch (timesketch.models.sketch.Sketch): a sketch object.

    Returns:
        pd.DataFrame: a pandas DataFrame with the results from
            the query.
    """
    lines = []
    for event in result['hits']['hits']:
        line = event['_source']
        line.setdefault('label', [])
        line['_id'] = event['_id']
        line['_type'] = event['_type']
        line['_index'] = event['_index']
        if 'tag' in line:
            if isinstance(line['tag'], (list, tuple)):
                line['tag'] = ','.join(line['tag'])
        try:
            for label in line['timesketch_label']:
                if sketch.id != label['sketch_id']:
                    continue
                line['label'].append(label['name'])
            del line['timesketch_label']
        except KeyError:
            pass

        lines.append(line)
    data_frame = pd.DataFrame(lines)
    del lines
    return data_frame


def query_results_to_filehandle(result, sketch):
    """Returns a data frame from a Elastic query result dict.

    Args:
        result (dict): a dict that contains the response from a
            Elastic datastore search.
        sketch (timesketch.models.sketch.Sketch): a sketch object.

    Returns:
        pd.DataFrame: a pandas DataFrame with the results from
            the query.
    """
    fh = io.StringIO()
    data_frame = query_results_to_dataframe(result, sketch)
    data_frame.to_csv(fh, index=False)
    fh.seek(0)
    return fh



# Copyright 2020 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""View resources for version 1 of the Timesketch API."""

import json

from flask import jsonify
from flask import request
from flask import abort
from flask_restful import Resource
from flask_login import login_required
from flask_login import current_user

from timesketch.api.v1 import resources
from timesketch.lib import forms
from timesketch.lib.definitions import HTTP_STATUS_CODE_OK
from timesketch.lib.definitions import HTTP_STATUS_CODE_CREATED
from timesketch.lib.definitions import HTTP_STATUS_CODE_BAD_REQUEST
from timesketch.lib.definitions import HTTP_STATUS_CODE_FORBIDDEN
from timesketch.lib.definitions import HTTP_STATUS_CODE_NOT_FOUND
from timesketch.models import db_session
from timesketch.models.sketch import Sketch
from timesketch.models.sketch import SearchTemplate
from timesketch.models.sketch import View


class ViewListResource(resources.ResourceMixin, Resource):
    """Resource to create a View."""

    @staticmethod
    def create_view_from_form(sketch, form):
        """Creates a view from form data.

        Args:
            sketch: Instance of timesketch.models.sketch.Sketch
            form: Instance of timesketch.lib.forms.SaveViewForm

        Returns:
            A view (Instance of timesketch.models.sketch.View)
        """
        # Default to user supplied data
        view_name = form.name.data
        query_string = form.query.data

        query_filter_dict = form.filter.data
        # Stripping potential pagination from views before saving it.
        if 'from' in query_filter_dict:
            del query_filter_dict['from']
        query_filter = json.dumps(query_filter_dict, ensure_ascii=False)
        query_dsl = json.dumps(form.dsl.data, ensure_ascii=False)

        if isinstance(query_filter, tuple):
            query_filter = query_filter[0]

        # No search template by default (before we know if the user want to
        # create a template or use an existing template when creating the view)
        searchtemplate = None

        # Create view from a search template
        if form.from_searchtemplate_id.data:
            # Get the template from the datastore
            template_id = form.from_searchtemplate_id.data
            searchtemplate = SearchTemplate.query.get(template_id)

            # Copy values from the template
            view_name = searchtemplate.name
            query_string = searchtemplate.query_string
            query_filter = searchtemplate.query_filter
            query_dsl = searchtemplate.query_dsl
            # WTF form returns a tuple for the filter. This is not
            # compatible with SQLAlchemy.
            if isinstance(query_filter, tuple):
                query_filter = query_filter[0]

        # Create a new search template based on this view (only if requested by
        # the user).
        if form.new_searchtemplate.data:
            query_filter_dict = json.loads(query_filter)
            if query_filter_dict.get('indices', None):
                query_filter_dict['indices'] = '_all'

            query_filter = json.dumps(query_filter_dict, ensure_ascii=False)

            searchtemplate = SearchTemplate(
                name=view_name,
                user=current_user,
                query_string=query_string,
                query_filter=query_filter,
                query_dsl=query_dsl)
            db_session.add(searchtemplate)
            db_session.commit()

        # Create the view in the database
        view = View(
            name=view_name,
            sketch=sketch,
            user=current_user,
            query_string=query_string,
            query_filter=query_filter,
            query_dsl=query_dsl,
            searchtemplate=searchtemplate)
        db_session.add(view)
        db_session.commit()

        return view

    @login_required
    def get(self, sketch_id):
        """Handles GET request to the resource.

        Args:
            sketch_id: Integer primary key for a sketch database model

        Returns:
            Views in JSON (instance of flask.wrappers.Response)
        """
        sketch = Sketch.query.get_with_acl(sketch_id)
        if not sketch:
            abort(
                HTTP_STATUS_CODE_NOT_FOUND, 'No sketch found with this ID.')
        if not sketch.has_permission(current_user, 'read'):
            abort(HTTP_STATUS_CODE_FORBIDDEN,
                  'User does not have read access controls on sketch.')
        return self.to_json(sketch.get_named_views)

    @login_required
    def post(self, sketch_id):
        """Handles POST request to the resource.

        Args:
            sketch_id: Integer primary key for a sketch database model

        Returns:
            A view in JSON (instance of flask.wrappers.Response)
        """
        form = forms.SaveViewForm.build(request)
        if not form.validate_on_submit():
            abort(
                HTTP_STATUS_CODE_BAD_REQUEST,
                'Unable to save view, not able to validate form data.')
        sketch = Sketch.query.get_with_acl(sketch_id)
        if not sketch:
            abort(
                HTTP_STATUS_CODE_NOT_FOUND, 'No sketch found with this ID.')
        if not sketch.has_permission(current_user, 'write'):
            abort(HTTP_STATUS_CODE_FORBIDDEN,
                  'User does not have write access controls on sketch.')
        view = self.create_view_from_form(sketch, form)
        return self.to_json(view, status_code=HTTP_STATUS_CODE_CREATED)


class ViewResource(resources.ResourceMixin, Resource):
    """Resource to get a view."""

    @login_required
    def get(self, sketch_id, view_id):
        """Handles GET request to the resource.

        Args:
            sketch_id: Integer primary key for a sketch database model
            view_id: Integer primary key for a view database model

        Returns:
            A view in JSON (instance of flask.wrappers.Response)
        """
        sketch = Sketch.query.get_with_acl(sketch_id)
        if not sketch:
            abort(
                HTTP_STATUS_CODE_NOT_FOUND, 'No sketch found with this ID.')
        view = View.query.get(view_id)

        if not sketch.has_permission(current_user, 'read'):
            abort(HTTP_STATUS_CODE_FORBIDDEN,
                  'User does not have read access controls on sketch.')

        # Check that this view belongs to the sketch
        if view.sketch_id != sketch.id:
            abort(
                HTTP_STATUS_CODE_NOT_FOUND,
                'Sketch ID ({0:d}) does not match with the sketch ID '
                'that is defined in the view ({1:d})'.format(
                    view.sketch_id, sketch.id))

        # If this is a user state view, check that it
        # belongs to the current_user
        if view.name == '' and view.user != current_user:
            abort(
                HTTP_STATUS_CODE_FORBIDDEN,
                'Unable to view a state view that belongs to a '
                'different user.')

        # Check if view has been deleted
        if view.get_status.status == 'deleted':
            meta = dict(deleted=True, name=view.name)
            schema = dict(meta=meta, objects=[])
            return jsonify(schema)

        # Make sure we have all expected attributes in the query filter.
        view.query_filter = view.validate_filter()
        db_session.add(view)
        db_session.commit()

        return self.to_json(view)

    @login_required
    def delete(self, sketch_id, view_id):
        """Handles DELETE request to the resource.

        Args:
            sketch_id: Integer primary key for a sketch database model
            view_id: Integer primary key for a view database model
        """
        sketch = Sketch.query.get_with_acl(sketch_id)
        if not sketch:
            abort(
                HTTP_STATUS_CODE_NOT_FOUND, 'No sketch found with this ID.')

        if not sketch.has_permission(current_user, 'write'):
            abort(HTTP_STATUS_CODE_FORBIDDEN,
                  'User does not have write access controls on sketch.')
        view = View.query.get(view_id)
        if not view:
            abort(
                HTTP_STATUS_CODE_NOT_FOUND, 'No view found with this ID.')

        # Check that this view belongs to the sketch
        if view.sketch_id != sketch.id:
            abort(
                HTTP_STATUS_CODE_NOT_FOUND,
                'The view does not belong to the sketch ({0:d} vs '
                '{1:d})'.format(view.sketch_id, sketch.id))

        if not sketch.has_permission(user=current_user, permission='write'):
            abort(
                HTTP_STATUS_CODE_FORBIDDEN,
                'User does not have write permission on sketch.')

        view.set_status(status='deleted')
        return HTTP_STATUS_CODE_OK

    @login_required
    def post(self, sketch_id, view_id):
        """Handles POST request to the resource.

        Args:
            sketch_id: Integer primary key for a sketch database model
            view_id: Integer primary key for a view database model

        Returns:
            A view in JSON (instance of flask.wrappers.Response)
        """
        form = forms.SaveViewForm.build(request)
        if not form.validate_on_submit():
            abort(
                HTTP_STATUS_CODE_BAD_REQUEST,
                'Unable to update view, not able to validate form data')
        sketch = Sketch.query.get_with_acl(sketch_id)
        if not sketch:
            abort(
                HTTP_STATUS_CODE_NOT_FOUND, 'No sketch found with this ID.')
        if not sketch.has_permission(current_user, 'write'):
            abort(HTTP_STATUS_CODE_FORBIDDEN,
                  'User does not have write access controls on sketch.')
        view = View.query.get(view_id)
        view.query_string = form.query.data

        query_filter = form.filter.data
        # Stripping potential pagination from views before saving it.
        if 'from' in query_filter:
            del query_filter['from']
        view.query_filter = json.dumps(query_filter, ensure_ascii=False)

        view.query_dsl = json.dumps(form.dsl.data, ensure_ascii=False)
        view.user = current_user
        view.sketch = sketch

        if form.dsl.data:
            view.query_string = ''

        db_session.add(view)
        db_session.commit()
        return self.to_json(view, status_code=HTTP_STATUS_CODE_CREATED)



"""Tests for WinCrashPlugin."""
from __future__ import unicode_literals

import unittest
import mock

from timesketch.lib.analyzers import win_crash
from timesketch.lib.testlib import BaseTest
from timesketch.lib.testlib import MockDataStore

class TestWinCrashPlugin(BaseTest):
    """Tests the functionality of the analyzer."""

    @mock.patch(
        'timesketch.lib.analyzers.interface.ElasticsearchDataStore',
        MockDataStore)
    def setUp(self):
        super().setUp()
        self.analyzer = win_crash.WinCrashSketchPlugin('test_index', 1)

    def test_formulate_query(self):
        """Test generator of Elasticsearch queries"""
        template = {
            'Event - App Error' : (
                'data_type:"windows:evtx:record"',
                'event_level:"2"',
            ),
            'File - WER Report' : (
                'data_type:"fs:stat"',
                'filename:"/Microsoft/Windows/WER/"',
            ),
            'Registry - Crash Reporting' : (
                'data_type:"windows:registry:key_value"',
                r'key_path:"\\Control\\CrashControl"',
            ),
        }
        expected_query = (
            '(data_type:"windows:evtx:record" AND event_level:"2")'
            ' OR (data_type:"fs:stat" AND filename:"/Microsoft/Windows/WER/")'
            ' OR (data_type:"windows:registry:key_value" AND '
            'key_path:"\\\\Control\\\\CrashControl")'
        ).replace('(', '').replace(')', '')

        # To prevent flakiness we break up the query into a set.
        expected_set = set(expected_query.split())

        formulated_query = self.analyzer.formulate_query(template)
        formulated_set = set(
            formulated_query.replace('(', '').replace(')', '').split())

        self.assertSetEqual(expected_set, formulated_set)

    def test_extract_filename(self):
        """Test generator of filename extraction regex"""
        string_list = (
            '\\WER\\ReportQueue\\AppCrash_notepad.exe_8d163e29d3960561ca2e972'
            '3640cd8fff5c2ad5_cab_0a5810ac\\Report.wer',
            '/WER/ReportQueue/NonCritical_notepad.exe_d473a376adfb18a7b165c5e'
            '3c26de43cd8bccb_cab_07fd2620',
            '/WER/ReportQueue/NonCritical_x64_d473a376adfb18a7b165c5e3c26de43'
            'cd8bccb_cab_07fd2620',
            'Strings: [\'iexplore.exe\', \'8.0.7601.17514\'')
        expected_list = (
            'notepad.exe',
            'notepad.exe',
            '',
            'iexplore.exe')

        for i, s in enumerate(string_list):
            self.assertEqual(
                self.analyzer.extract_filename(s),
                expected_list[i])

    def test_mark_as_crash(self):
        mock_event = mock.Mock()
        filename = 'sample_filename.exe'
        self.analyzer.mark_as_crash(mock_event, filename)
        mock_event.add_attributes.assert_called_once_with(
            {'crash_app': filename})
        mock_event.add_tags.assert_called_once_with(['win_crash'])

        mock_event = mock.Mock()
        self.analyzer.mark_as_crash(mock_event, None)
        mock_event.add_attributes.assert_not_called()
        mock_event.add_tags.assert_called_once_with(['win_crash'])

if __name__ == '__main__':
    unittest.main()



"""Tests for SSHSessionizerSketchPlugin"""

from __future__ import unicode_literals

import mock

from timesketch.lib.analyzers.ssh_sessionizer import SSHSessionizerSketchPlugin
from timesketch.lib.testlib import BaseTest
from timesketch.lib.testlib import MockDataStore
# TODO _create_mock_event will be renamed in another pull request. It's name
# should be also changed here.
from timesketch.lib.analyzers.sequence_sessionizer_test \
    import _create_mock_event

# Message attributes for events that represent one mock SSH session.
one_ssh_session_args = [{
    'message':
    '[sshd] [1]: Connection from 1.1.1.1 port 1 on 1.1.1.1 port 1'
}, {
    'message': '[sshd] [1]: Accepted certificate ID'
}]

# Message attributes for events that represent two mock SSH sessions.
many_ssh_session_args = [{
    'message':
    '[sshd] [1]: Connection from 1.1.1.1 port 1 on 1.1.1.1 port 1'
}, {
    'message': '[sshd] [1]: Accepted certificate ID'
}, {
    'message':
    '[sshd] [2]: Connection from 2.2.2.2 port 2 on 2.2.2.2 port 2'
}, {
    'message': '[sshd] [2]: Accepted certificate ID'
}]

# Message attributes for a SSH event that is not a connection SSH event
no_ssh_session_args = [{
    'message': '[sshd] [0]: Loaded keys'
}]


class TestSSHSessionizerPlugin(BaseTest):
    """Tests the functionality of the ssh sessionizing sketch analyzer."""
    @mock.patch('timesketch.lib.analyzers.interface.ElasticsearchDataStore',
                MockDataStore)
    def test_sessionizer(self):
        """Test basic ssh sessionizer functionality."""
        index = 'test_index'
        sketch_id = 1
        sessionizer = SSHSessionizerSketchPlugin(index, sketch_id)
        self.assertIsInstance(sessionizer, SSHSessionizerSketchPlugin)
        self.assertEqual(index, sessionizer.index_name)
        self.assertEqual(sketch_id, sessionizer.sketch.id)

    @mock.patch('timesketch.lib.analyzers.interface.ElasticsearchDataStore',
                MockDataStore)
    def test_session_starts_with_connection_event(self):
        """Test a session is created if it starts with SSH connection event."""
        index = 'test_index'
        sketch_id = 1
        sessionizer = SSHSessionizerSketchPlugin(index, sketch_id)
        sessionizer.datastore.client = mock.Mock()
        datastore = sessionizer.datastore

        _create_mock_event(datastore, 0, 1, one_ssh_session_args)

        message = sessionizer.run()
        self.assertEqual(
            message,
            'Sessionizing completed, number of ssh_session sessions created: 1'
        )
        session_id = '1.1.1.1_1'
        event = datastore.event_store['0']
        self.assertEqual(event['_source']['session_id']['ssh_session'],
                         session_id)

    @mock.patch('timesketch.lib.analyzers.interface.ElasticsearchDataStore',
                MockDataStore)
    def test_all_events_from_session_are_labeled(self):
        """Test one SSH session of events is finded and allocated correctly."""
        index = 'test_index'
        sketch_id = 1
        sessionizer = SSHSessionizerSketchPlugin(index, sketch_id)
        sessionizer.datastore.client = mock.Mock()
        datastore = sessionizer.datastore

        _create_mock_event(datastore, 0, 2, one_ssh_session_args, [1])

        message = sessionizer.run()
        self.assertEqual(
            message,
            'Sessionizing completed, number of ssh_session sessions created: 1'
        )

        session_id = '1.1.1.1_1'
        event = datastore.event_store['0']
        self.assertEqual(event['_source']['session_id']['ssh_session'],
                         session_id)
        event = datastore.event_store['101']
        self.assertEqual(event['_source']['session_id']['ssh_session'],
                         session_id)

    @mock.patch('timesketch.lib.analyzers.interface.ElasticsearchDataStore',
                MockDataStore)
    def test_session_doesnt_start_with_no_connection_event(self):
        """Test a session is not created if it doesn't start with SSH connection
        event."""
        index = 'test_index'
        sketch_id = 1
        sessionizer = SSHSessionizerSketchPlugin(index, sketch_id)
        sessionizer.datastore.client = mock.Mock()
        datastore = sessionizer.datastore

        _create_mock_event(datastore, 0, 1, no_ssh_session_args)

        message = sessionizer.run()
        self.assertEqual(
            message,
            'Sessionizing completed, number of ssh_session sessions created: 0'
        )

        event = datastore.event_store['0']
        self.assertNotIn('session_id', event['_source'])

    @mock.patch('timesketch.lib.analyzers.interface.ElasticsearchDataStore',
                MockDataStore)
    def test_multiple_sessions(self):
        """Test multiple sessions are found and allocated correctly."""
        index = 'test_index'
        sketch_id = 1
        sessionizer = SSHSessionizerSketchPlugin(index, sketch_id)
        sessionizer.datastore.client = mock.Mock()
        datastore = sessionizer.datastore

        _create_mock_event(datastore,
                           0,
                           4,
                           many_ssh_session_args,
                           time_diffs=[1, 1, 1])

        message = sessionizer.run()
        self.assertEqual(
            message,
            'Sessionizing completed, number of ssh_session sessions created: 2'
        )

        session_id_1 = '1.1.1.1_1'
        session_id_2 = '2.2.2.2_2'

        event = datastore.event_store['0']
        self.assertEqual(event['_source']['session_id']['ssh_session'],
                         session_id_1)
        event = datastore.event_store['101']
        self.assertEqual(event['_source']['session_id']['ssh_session'],
                         session_id_1)
        event = datastore.event_store['202']
        self.assertEqual(event['_source']['session_id']['ssh_session'],
                         session_id_2)
        event = datastore.event_store['303']
        self.assertEqual(event['_source']['session_id']['ssh_session'],
                         session_id_2)



"""Sketch analyzer plugin for the Safe Browsing API."""
from __future__ import unicode_literals

import fnmatch
import logging
import re

import pkg_resources
import requests
from flask import current_app

from timesketch.lib.analyzers import interface
from timesketch.lib.analyzers import manager


logger = logging.getLogger('timesketch.tasks')


class SafeBrowsingSketchPlugin(interface.BaseSketchAnalyzer):
    """Sketch analyzer for Safe Browsing."""

    NAME = 'safebrowsing'

    # Safe Browsing API v4, threatMatches.find endpoint.
    _SAFE_BROWSING_THREATMATCHING_ENDPOINT = (
        'https://safebrowsing.googleapis.com/v4/threatMatches:find')

    # Maximal number of URLs in a single Lookup request as per the
    # Safe Browsing API documentation.
    _SAFE_BROWSING_BULK_LIMIT = 500

    # An optional file containing URL wildcards to be allow listed
    # in a YAML file.
    _URL_ALLOW_LIST_CONFIG = 'safebrowsing_allowlist.yaml'

    # The keys to be added to the TS event from the ThreatMatch object
    # we get from Safe Browsing API.
    _SAFEBROWSING_ENTRY_KEEP = frozenset([
        'platformType',
        'threatType',
    ])

    # Used to find proper URLs in the 'url' entries of TS events.
    _URL_BEGINNING_RE = re.compile(r'(http(s|):\/\/\S*)')

    def __init__(self, index_name, sketch_id):
        """Initialize The Sketch Analyzer.

        Args:
            index_name: Elasticsearch index name
            sketch_id: Sketch ID
        """
        super(SafeBrowsingSketchPlugin, self).__init__(index_name, sketch_id)

        self._safebrowsing_api_key = current_app.config.get(
            'SAFEBROWSING_API_KEY')

        self._google_client_id = current_app.config.get(
            'SAFEBROWSING_CLIENT_ID',
            'Timesketch',
        )

        self._google_client_version = current_app.config.get(
            'SAFEBROWSING_CLIENT_VERSION',
            pkg_resources.get_distribution('timesketch').version,
        )

    def _is_url_allowlisted(self, url, allowlist):
        """Does a fnmatch against the allowlist.

        Args:
            url: The url
            allowlist: The allowlist, list-like
        Returns:
            Boolean with the result
        """

        for url_pattern in allowlist:
            if fnmatch.fnmatchcase(url, url_pattern):
                return True

        return False

    def _do_safebrowsing_lookup(self, urls, platforms, types):
        """URL lookup against the Safe Browsing API.

        Args:
            urls: URLs
            platforms: platformTypes field of threatInfo
            types: threatTypes field of threatInfo
        Returns:
            Dict of URLs with the hits
        """
        results = {}

        api_client = {
            'clientId': self._google_client_id,
            'clientVersion': self._google_client_version,
        }

        for index in range(0, len(urls), self._SAFE_BROWSING_BULK_LIMIT):
            body = {
                'client': api_client,
                'threatInfo': {
                    'platformTypes': platforms,
                    'threatTypes': types,
                    'threatEntryTypes': ['URL'],
                    'threatEntries': [
                        {'url': url}
                        for url in urls[
                            index:index + self._SAFE_BROWSING_BULK_LIMIT]
                    ],
                },
            }

            response = requests.post(
                self._SAFE_BROWSING_THREATMATCHING_ENDPOINT,
                params={'key': self._safebrowsing_api_key},
                json=body,
            )

            try:
                response.raise_for_status()
            except requests.exceptions.HTTPError as e:
                logger.error(e)
                continue

            result = response.json()

            if not result:
                continue

            if 'matches' not in result:
                continue

            for match in result.get('matches'):
                result_url = match.get('threat', {}).get('url')

                if not result_url:
                    continue

                # Removing all key/values that are not defined in
                # the _SAFEBROWSING_ENTRY_KEEP.
                threat_result = match.copy()
                for key in match.keys():
                    if key not in self._SAFEBROWSING_ENTRY_KEEP:
                        threat_result.pop(key)

                results[result_url] = threat_result

        return results

    def _sanitize_url(self, url_entry):
        """Finds http[s]:// in 'url_entry' and returns its content from there.

        Args:
            url_entry: a URL, with some other characters before and after

        Returns:
            String with the URL or empty string if not found
        """
        m = self._URL_BEGINNING_RE.search(url_entry)

        if m:
            return m.group(1)

        return ''

    def run(self):
        """Entry point for the analyzer.

        Returns:
            String with summary of the analyzer result
        """
        # Exit ASAP if the API key is missing.
        if not self._safebrowsing_api_key:
            return 'Safe Browsing API requires an API key!'

        query = (
            '{"query": { "bool": { "should": [ '
            '{ "exists" : { "field" : "url" }} ] } } }')

        return_fields = ['url']

        events = self.event_stream(
            query_dsl=query,
            return_fields=return_fields,
        )

        urls = {}

        for event in events:
            url = self._sanitize_url(event.source.get('url'))

            if not url:
                continue

            urls.setdefault(url, []).append(event)

        # Exit early if there are no URLs in the data set to analyze.
        if not urls:
            return 'No URLs to analyze.'

        url_allowlisted = 0

        url_allowlist = set(
            interface.get_yaml_config(
                self._URL_ALLOW_LIST_CONFIG,
            ),
        )

        if not url_allowlist:
            domain_analyzer_allowlisted = current_app.config.get(
                'DOMAIN_ANALYZER_EXCLUDE_DOMAINS',
                [],
            )
            for domain in domain_analyzer_allowlisted:
                url_allowlist.add('*.%s/*' % domain)

        logger.info(
            '{0:d} entries on the allowlist.'.format(len(url_allowlist)),
        )

        safebrowsing_platforms = current_app.config.get(
            'SAFEBROWSING_PLATFORMS',
            ['ANY_PLATFORM'],
        )

        safebrowsing_types = current_app.config.get(
            'SAFEBROWSING_THREATTYPES',
            ['MALWARE'],
        )

        lookup_urls = []

        for url in urls:
            if self._is_url_allowlisted(url, url_allowlist):
                url_allowlisted += 1
                continue

            lookup_urls.append(url)

        try:
            safebrowsing_results = self._do_safebrowsing_lookup(
                lookup_urls,
                safebrowsing_platforms,
                safebrowsing_types,
            )
        except requests.HTTPError:
            return 'Couldn\'t reach the Safe Browsing API.'

        for url in lookup_urls:
            safebrowsing_result = safebrowsing_results.get(url)

            if not safebrowsing_result:
                continue

            for event in urls[url]:
                tags = ['google-safebrowsing-url']

                threat_type = safebrowsing_result.get('threatType')

                if threat_type:
                    tags.append(
                        'google-safebrowsing-%s' % threat_type.lower(),
                    )

                event.add_tags(tags)

                threat_attributes = []
                for item in safebrowsing_result.items():
                    threat_attributes.append('%s: %s' % item)

                event.add_attributes(
                    {
                        'google-safebrowsing-threat': ', '.join(
                            threat_attributes),
                    },
                )
                event.commit()

        return (
            '{0:d} Safe Browsing result(s) on {1:d} URL(s), '
            '{2:d} on the allow list.').format(
                len(safebrowsing_results),
                len(urls),
                url_allowlisted,
            )


manager.AnalysisManager.register_analyzer(SafeBrowsingSketchPlugin)



# Copyright 2018 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Based on the example code from the public Google IAP documentation:
https://cloud.google.com/iap/docs/signed-headers-howto
"""

from __future__ import unicode_literals

import time
import json
import hashlib
import os
import six

# six.moves is a dynamically-created namespace that doesn't actually
# exist and therefore pylint can't statically analyze it.
# pylint: disable-msg=import-error
from six.moves.urllib import parse as urlparse

import jwt
import requests

from flask import url_for
from flask import current_app
from flask import session

from timesketch.lib.definitions import HTTP_STATUS_CODE_OK


CSRF_KEY = 'google_oauth2_csrf_token'
AUTH_URI = 'https://accounts.google.com/o/oauth2/v2/auth'
DISCOVERY_URL = 'https://accounts.google.com/.well-known/openid-configuration'


class JwtValidationError(Exception):
    """Raised when a JSON Web Token cannot be validated."""


class JwtKeyError(Exception):
    """Raised when there is a problem with the public key used for signing."""


class JwtFetchError(Exception):
    """Raised when there is a problem with the public key used for signing."""


class DiscoveryDocumentError(Exception):
    """Raised when there is a problem with the discovery document."""


def _fetch_public_keys(url):
    """Fetch public keys used for verifying signatures.

    Args:
        url: URL where keys can be fetched.

    Raises:
        JwTKeyError if keys cannot be fetched.

    Returns:
        HTTP response.
    """
    try:
        resp = requests.get(url)
    except requests.exceptions.RequestException as e:
        raise JwtKeyError('Cannot fetch public keys: {}'.format(e))
    if resp.status_code != HTTP_STATUS_CODE_OK:
        raise JwtKeyError(
            'Cannot fetch public keys: {}'.format(resp.status_code))
    return resp.json()


def _fetch_oauth2_discovery_document():
    """Fetch Google OAuth2 discovery document.

    Raises:
        DiscoveryDocumentError if document cannot be fetched.

    Returns:
        HTTP response.
    """
    try:
        resp = requests.get(DISCOVERY_URL)
    except requests.exceptions.RequestException as e:
        raise DiscoveryDocumentError(
            'Cannot fetch discovery document: {}'.format(e))
    if resp.status_code != HTTP_STATUS_CODE_OK:
        raise DiscoveryDocumentError(
            'Cannot fetch discovery_document: {}'.format(resp.status_code))
    return resp.json()


def _generate_random_token():
    """Generate random string to use as CSRF and nonce tokens.

    Returns:
        Random string.
    """
    return hashlib.sha256(os.urandom(1024)).hexdigest()


def get_oauth2_authorize_url(hosted_domain=None):
    """Generate an authorization URL for Google's OAuth2 service.

    Args:
        hosted_domain: Optional GSuite domain to limit access to.

    Returns:
        Authorization URL.
    """
    csrf_token = _generate_random_token()
    nonce = _generate_random_token()
    redirect_uri = url_for(
        'user_views.google_openid_connect',
        _scheme='https',
        _external=True
    )
    scopes = ('openid', 'email', 'profile')

    # Add the generated CSRF token to the client session for later validation.
    session[CSRF_KEY] = csrf_token

    # Generate authorization URL
    params = dict(
        client_id=current_app.config.get('GOOGLE_OIDC_CLIENT_ID'),
        scope=' '.join(scopes),
        response_type='code',
        access_type='online',  # Because we don't need a refresh token.
        state=csrf_token,
        nonce=nonce,  # Enable replay attack protection attack.
        redirect_uri=redirect_uri
    )
    if hosted_domain:
        params['hd'] = hosted_domain

    urlencoded_params = urlparse.urlencode(params)
    google_authorization_url = '{}?{}'.format(AUTH_URI, urlencoded_params)
    return google_authorization_url


def get_encoded_jwt_over_https(code):
    """Fetch a JSON Web Token (JWT) using a authentication code.

    Args:
        code: Authentication code obtained from an OAuth2 flow.

    Raises:
        JwtFetchError if JWT cannot be fetched.

    Returns:
        Encoded JWT.
    """

    discovery_document = get_oauth2_discovery_document()
    redirect_uri = url_for(
        'user_views.google_openid_connect',
        _scheme='https',
        _external=True
    )
    post_data = {
        'code': code,
        'client_id': current_app.config.get('GOOGLE_OIDC_CLIENT_ID'),
        'client_secret': current_app.config.get('GOOGLE_OIDC_CLIENT_SECRET'),
        'redirect_uri': redirect_uri,
        'grant_type': 'authorization_code'
    }
    token_url = discovery_document.get('token_endpoint')
    try:
        response = requests.post(token_url, data=post_data)
        encoded_jwt = response.json().get('id_token')
    except requests.exceptions.RequestException as e:
        raise JwtFetchError(
            'Cannot fetch JWT: {}'.format(e))
    if response.status_code != HTTP_STATUS_CODE_OK:
        raise JwtFetchError(
            'Cannot fetch JWT: {}'.format(response.status_code))

    if not encoded_jwt:
        raise JwtFetchError('Cannot fetch JWT: Missing id_token in response')

    return encoded_jwt


def decode_jwt(encoded_jwt, public_key, algorithm, expected_audience):
    """Decode a JSON Web Token (JWT).

    Args:
        encoded_jwt: The contents of the X-Goog-IAP-JWT-Assertion header.
        public_key: Key to verify signature of the JWT.
        algorithm: Algorithm used for the key. E.g. ES256, RS256
        expected_audience: Expected audience in the JWT.

    Returns:
        Decoded JWT as a dict object.

    Raises:
        JwtValidationError: if the JWT token cannot be decoded.
    """
    try:
        decoded_jwt = jwt.decode(
            encoded_jwt, public_key, algorithm=algorithm,
            audience=expected_audience)
        return decoded_jwt
    except (jwt.exceptions.InvalidTokenError,
            jwt.exceptions.InvalidKeyError) as e:
        raise JwtValidationError('JWT validation error: {}'.format(e))

    return None


def validate_jwt(decoded_jwt, expected_issuer, expected_domain=None):
    """Decode and validate a JSON Web token (JWT).

    Cloud IAP:
    https://cloud.google.com/iap/docs/signed-headers-howto

    Google OpenID Connect:
    https://developers.google.com/identity/protocols/OpenIDConnect

    Args:
        decoded_jwt: A dict object containing the decoded JWT token.
        expected_issuer: Expected issuer of the JWT.
        expected_domain: Expected GSuite domain in the JWT (optional).

    Raises:
        JwtValidationError: If unable to validate the JWT.
    """
    # Make sure the token is not created in the future or has expired.
    try:
        now = int(time.time())
        issued_at = decoded_jwt['iat']
        if isinstance(issued_at, six.string_types):
            issued_at = int(issued_at, 10)

        expires_at = decoded_jwt['exp']
        if isinstance(expires_at, six.string_types):
            expires_at = int(expires_at, 10)

        if issued_at > now:
            raise JwtValidationError('Token was issued in the future')
        if expires_at < now:
            raise JwtValidationError('Token has expired')
    except KeyError as e:
        raise JwtValidationError('Missing timestamp: {}'.format(e))

    # Check that the issuer of the token is correct.
    try:
        issuer = decoded_jwt['iss']
        if issuer != expected_issuer:
            raise JwtValidationError('Wrong issuer: {}'.format(issuer))
    except KeyError:
        raise JwtValidationError('Missing issuer')

    if 'email' not in decoded_jwt:
        raise JwtValidationError('Missing email field in token')

    if expected_domain:
        try:
            domain = decoded_jwt['hd']
            if not domain == expected_domain:
                raise JwtValidationError('Wrong domain: {}'.format(domain))
        except KeyError as e:
            raise JwtValidationError('Missing domain: {}'.format(e))


def get_public_key_for_jwt(encoded_jwt, url):
    """Get public key for JWT in order to verify the signature.

    The keys get cached in order to limit the amount of network round trips.

    Args:
        encoded_jwt: Base64 encoded JWT.
        url: URL where keys can be fetched.

    Raises:
        JwTKeyError if keys cannot be fetched.

    Returns:
        Key as string.
    """
    # Get the Key ID from the JWT header.
    key_id = jwt.get_unverified_header(encoded_jwt).get('kid')
    if not key_id:
        raise JwtKeyError('Missing key ID field in token header')
    key_cache = get_public_key_for_jwt.key_cache
    key = key_cache.get(key_id)
    if key:
        return key

    # Re-fetch the key file.
    keys_json = _fetch_public_keys(url)
    if 'keys' in keys_json:
        _new_keys_dict = {}
        for key_dict in keys_json['keys']:
            public_key = jwt.algorithms.RSAAlgorithm.from_jwk(
                json.dumps(key_dict))
            _new_keys_dict[key_dict['kid']] = public_key
        key_cache = _new_keys_dict
    else:
        key_cache = keys_json
    get_public_key_for_jwt.key_cache = key_cache
    key = key_cache.get(key_id)
    if not key:
        raise JwtKeyError('IAP public key {!r} not found'.format(key_id))

    return key


def get_oauth2_discovery_document():
    """Get Google OAuth2 discovery document.

    The document is cached in order to limit the amount of network round trips
    and is set to expire in 12 hours from when it was fetched.

    Returns:
        Discovery document as a dictionary.
    """
    now = int(time.time())
    cache = get_oauth2_discovery_document.cache
    discovery_document = cache.get('current')
    if discovery_document:
        # Check if the document has expired.
        created_at = discovery_document['created_at']
        expires_at = created_at + 12*60*60  # 12 hours in seconds
        if now < expires_at:
            return discovery_document['document']

    # Re-fetch the discovery document.
    new_discovery_document = _fetch_oauth2_discovery_document()
    cache = {
        'current': {
            'created_at': now,
            'document': new_discovery_document
        }
    }
    get_oauth2_discovery_document.cache = cache
    return new_discovery_document


# Used to cache public keys.
get_public_key_for_jwt.key_cache = {}

# Used to cache discovery documents.
get_oauth2_discovery_document.cache = {}



"""Sketch analyzer plugin for browser timeframe."""
from __future__ import unicode_literals

import collections
import pandas as pd

from timesketch.lib import emojis
from timesketch.lib.analyzers import interface
from timesketch.lib.analyzers import manager
from timesketch.lib.analyzers import utils


def get_list_of_consecutive_sequences(hour_list):
    """Returns a list of runs from a list of numbers.

    Args:
        hour_list: a list of integers.

    Returns:
        A list of tuples, where each tuple indicates the first
        and last record of a consecutive run of numbers, eg
        list 1, 2, 3, 7, 8, 9 would produce the output
        of (1,3), (7, 9).
    """
    runs = []
    if not hour_list:
        return runs

    start = hour_list[0]
    now = start
    for hour in hour_list[1:]:
        if hour == (now + 1):
            now = hour
            continue

        runs.append((start, now))
        start = hour
        now = start

    if not runs:
        runs = [(hour_list[0], hour_list[-1])]

    last_start, last_end = runs[-1]

    if (start != last_start) and (last_end != now):
        runs.append((start, now))

    return runs


def fix_gap_in_list(hour_list):
    """Returns a list with gaps in it fixed.

    Args:
        hour_list: a list of integers in a sequence, potentially
            with holes in the sequence.

    Returns:
        A list that consists of the input numbers with single
        integer gaps filled. The list should not have more than
        two runs. Therefore if there are more than two runs after
        all gaps have been filled the "extra" runs will be dropped.
    """
    if not hour_list:
        return hour_list

    runs = get_list_of_consecutive_sequences(hour_list)
    len_runs = len(runs)

    for i in range(0, len_runs - 1):
        _, upper = runs[i]
        next_lower, _ = runs[i+1]
        if (upper + 1) == (next_lower - 1):
            hour_list.append(upper + 1)

    hours = sorted(hour_list)
    runs = get_list_of_consecutive_sequences(hour_list)

    if len(runs) <= 2:
        return hours
    if len_runs < len(runs):
        return fix_gap_in_list(hours)

    # Now we need to remove runs, we only need the first and last.
    run_start = runs[0]
    run_end = runs[-1]
    hours = list(range(0, run_start[1] + 1))
    hours.extend(range(run_end[0], run_end[1] + 1))

    return sorted(hours)


def get_active_hours(frame):
    """Return a list of the hours with the most activity within a frame.

    Args:
        frame: a pandas DataFrame object that contains a datetime column.

    Returns:
        A tuple that contains three items:
            1. list of hours where the most activity within the DataFrame
            occurs.
            2. the threshold used to determine what is considered to be
            an active hour.
            3. a DataFrame object containing the aggregation.

    """
    frame_count = frame[['datetime', 'hour']].groupby(
        'hour', as_index=False).count()
    frame_count['count'] = frame_count['datetime']
    del frame_count['datetime']

    stats = frame_count['count'].describe()

    # Define few different options for the threshold value of what constitutes
    # an active hour. Then we choose the method that has the highest active
    # hour count, as long as it is between 3 and 12 hours.
    thresholds = {
        stats['75%'] - stats['mean']: 0,
        stats['75%']: 0,
        stats['50%']: 0,
        stats['25%']: 0
    }
    for threshold in thresholds:
        threshold_filter = frame_count['count'] >= threshold
        hours = list(frame_count[threshold_filter].hour.values)
        hours = sorted(hours)

        hour_len = len(hours)
        if 3 <= hour_len <= 12:
            thresholds[threshold] = hour_len

    threshold_counter = collections.Counter(thresholds)
    threshold, _ = threshold_counter.most_common(1)[0]

    threshold_filter = frame_count['count'] >= threshold
    hours = list(frame_count[threshold_filter].hour.values)
    hours = sorted(hours)
    runs = get_list_of_consecutive_sequences(hours)

    # There should either be a single run or at most two.
    number_runs = len(runs)
    if number_runs == 1:
        return hours, threshold, frame_count

    if number_runs == 2 and runs[0][0] == 0:
        # Two runs, first one starts at hour zero.
        return hours, threshold, frame_count

    return fix_gap_in_list(hours), threshold, frame_count


class BrowserTimeframeSketchPlugin(interface.BaseSketchAnalyzer):
    """Sketch analyzer for BrowserTimeframe."""

    NAME = 'browser_timeframe'
    DEPENDENCIES = frozenset()

    def __init__(self, index_name, sketch_id):
        """Initialize The Sketch Analyzer.

        Args:
            index_name: Elasticsearch index name
            sketch_id: Sketch ID
        """
        self.index_name = index_name
        super(BrowserTimeframeSketchPlugin, self).__init__(
            index_name, sketch_id)

    def run(self):
        """Entry point for the analyzer.

        Returns:
            String with summary of the analyzer result
        """
        # TODO: Once we can identify user generated events this should be
        # updated to include all user generated events instead of focusing
        # solely on browser events.
        query = 'source_short:"WEBHIST" OR source:"WEBHIST"'

        return_fields = ['datetime', 'timestamp', 'url', 'tag', '__ts_emojis']

        data_frame = self.event_pandas(
            query_string=query, return_fields=return_fields)

        if not data_frame.shape[0]:
            return 'No browser events discovered.'

        sleeping_emoji = emojis.get_emoji('SLEEPING_FACE')

        # This query filters out all timestamps that have a zero timestamp as
        # well as those that occur after 2038-01-01, this may need to be
        # changed in the future.
        data_frame['timestamp'] = pd.to_numeric(data_frame.timestamp)
        data_frame = data_frame[
            (data_frame.timestamp > 0) & (
                data_frame.timestamp < 2145916800000000)]

        data_frame['datetime'] = pd.to_datetime(
            data_frame.timestamp / 1e6, utc=True, unit='s')
        data_frame['hour'] = pd.to_numeric(
            data_frame.datetime.dt.strftime('%H'))

        total_count = data_frame.shape[0]
        activity_hours, threshold, aggregation = get_active_hours(data_frame)

        if not activity_hours:
            return 'Did not discover any activity hours.'

        hour_count = dict(aggregation.values.tolist())
        data_frame_outside = data_frame[~data_frame.hour.isin(activity_hours)]

        for event in utils.get_events_from_data_frame(
                data_frame_outside, self.datastore):
            event.add_tags(['outside-active-hours'])
            hour = event.source.get('hour')
            this_hour_count = hour_count.get(hour)
            event.add_attributes(
                {'activity_summary': (
                    'Number of events for this hour ({0:d}): {1:d}, with the '
                    'threshold value: {2:0.2f}').format(
                        hour, this_hour_count, threshold),
                 'hour_count': this_hour_count})
            event.add_emojis([sleeping_emoji])
            event.commit()

        tagged_events, _ = data_frame_outside.shape
        if tagged_events:
            story = self.sketch.add_story('{0:s} - {1:s}'.format(
                utils.BROWSER_STORY_TITLE, self.timeline_name))
            story.add_text(
                utils.BROWSER_STORY_HEADER, skip_if_exists=True)

            # Find some statistics about the run time of the analyzer.
            percent = (tagged_events / total_count) * 100.0
            last_hour = activity_hours[0]
            end = 0
            for hour in activity_hours[1:]:
                if hour != last_hour + 1:
                    end = hour
                    break
                last_hour = hour

            if not end:
                first = activity_hours[0]
                last = activity_hours[-1]
            else:
                first = end
                index = activity_hours.index(end)
                last = activity_hours[index - 1]

            story.add_text(
                '## Browser Timeframe Analyzer\n\nThe browser timeframe '
                'analyzer discovered {0:d} browser events that occurred '
                'outside of the typical browsing window of this browser '
                'history ({1:s}), or around {2:0.2f}%  of the {3:d} total '
                'events.\n\nThe analyzer determines the activity hours by '
                'finding the frequency of browsing events per hour, and then '
                'discovering the longest block of most active hours before '
                'proceeding with flagging all events outside of that time '
                'period. This information can be used by other analyzers '
                'or by manually looking for other activity within the '
                'inactive time period to find unusual actions.\n\n'
                'The hours considered to be active hours are the hours '
                'between {4:02d} and {5:02d} (hours in UTC) and the '
                'threshold used to determine if an hour was considered to be '
                'active was: {6:0.2f}.'.format(
                    tagged_events, self.timeline_name, percent, total_count,
                    first, last, threshold))

            group = self.sketch.add_aggregation_group(
                name='Browser Activity Per Hour',
                description='Created by the browser timeframe analyzer')
            group.set_layered()

            params = {
                'data': aggregation.to_dict(orient='records'),
                'title': 'Browser Activity Per Hour ({0:s})'.format(
                    self.timeline_name),
                'field': 'hour',
                'order_field': 'hour',
            }
            agg_obj = self.sketch.add_aggregation(
                name='Browser Activity Per Hour ({0:s})'.format(
                    self.timeline_name), agg_name='manual_feed',
                agg_params=params, chart_type='barchart',
                description='Created by the browser timeframe analyzer',
                label='informational')
            group.add_aggregation(agg_obj)

            lines = [{'hour': x, 'count': threshold} for x in range(0, 24)]
            params = {
                'data': lines,
                'title': 'Browser Timeframe Threshold ({0:s})'.format(
                    self.timeline_name),
                'field': 'hour',
                'order_field': 'hour',
                'chart_color': 'red',
            }
            agg_line = self.sketch.add_aggregation(
                name='Browser Activity Per Hour ({0:s})'.format(
                    self.timeline_name), agg_name='manual_feed',
                agg_params=params, chart_type='linechart',
                description='Created by the browser timeframe analyzer',
                label='informational')
            group.add_aggregation(agg_line)
            story.add_aggregation_group(group)

        return (
            'Tagged {0:d} out of {1:d} events as outside of normal '
            'active hours.').format(tagged_events, total_count)


manager.AnalysisManager.register_analyzer(BrowserTimeframeSketchPlugin)
